import typer
import time
import glob
from typing import Optional, Any
from fuxion.generators import *
from fuxion.normalizers import *
from typing import Any, List, ClassVar, Optional
from fuxion.base import BaseChain
from typing import ClassVar
from fuxion.few_shot import generate_population, populate_few_shot
import os
from fuxion.generators import GeneratorChain
from fuxion.normalizers import NormalizerChain


class DatasetPipeline(BaseChain):

    k: int = 10
    few_shot_file: str
    sample_size: int = 3
    batch_save: bool = False
    batch_size: int = 100
    dataset_name: Optional[str] = None
    manual_review: bool = False
    generator: ClassVar[GeneratorChain]
    normalizer: ClassVar[NormalizerChain]
    chain_type = "DatasetPipeline"

    @property
    def input_keys(self) -> List[str]:
        return []

    @property
    def output_keys(self) -> List[str]:
        return ["dataset"]

    @classmethod
    def from_template(
        cls,
        *args,
        **kwargs,
    ):
        return super()._from_name(
            *args,
            generator_chain=GeneratorChain,
            normalizer_chain=NormalizerChain,
            class_suffix="DatasetPipeline",
            base_cls=DatasetPipeline,
            **kwargs,
        )

    def execute(
        self,
    ) -> dict[str, List[dict[str, Any | str]]]:

        return self.run()

    def run(
        self, inputs: dict[str, str] | None = None
    ) -> dict[str, List[dict[str, Any | str]]]:
        """
        Executes the dataset generation pipeline. This method is the primary entry point for running
        the dataset generation and normalization process. It delegates to the _call method with the
        provided inputs, ensuring that inputs are always in the expected format for the generator and
        normalizer chains.

        Args:
            inputs (Optional[dict[str, str]]): Optional dictionary of inputs to pass to the generator. Defaults to None.

        Returns:
            dict[str, List[dict[str, Any | str]]]: The results of the dataset generation and normalization process.
        """

        return self._call(inputs)

    def save_batch(self, batch: list[dict[str, Any]], batch_index: int):
        """Save a batch of outputs to a file, with each batch saved in a separate file."""

        if self.dataset_name is None:
            self.dataset_name = str(int(time.time()))

        batch_name = f"{self.dataset_name}_batch_{batch_index}.json"
        batch_dir = os.path.join(os.path.dirname(__file__), "datasets")
        if not os.path.exists(batch_dir):
            os.makedirs(batch_dir, exist_ok=True)

        batch_path = os.path.join(batch_dir, batch_name)
        batch_outputs: dict[str, dict[str, list[dict[str, Any | str]] | str]] = {
            "dataset": {
                "outputs": batch,
                "generator_prompt": self.generator.template.template,
                "normalizer_prompt": self.normalizer.template.template,
            }
        }
        with open(batch_path, "w") as fd:
            json.dump(batch_outputs, fd)

    def save_full_dataset(
        self, results: dict[str, dict[str, list[dict[str, Any | str]] | str]]
    ) -> None:
        """Save the complete dataset to a file."""
        if self.dataset_name is None:
            self.dataset_name = f"{str(int(time.time()))}.json"
        else:
            self.dataset_name = f"{self.dataset_name}.json"

        dataset_dir = os.path.join(os.path.dirname(__file__), "datasets")

        if not os.path.exists(dataset_dir):
            os.makedirs(dataset_dir, exist_ok=True)

        dataset_path = os.path.join(dataset_dir, self.dataset_name)

        with open(dataset_path, "w") as fd:
            json.dump(results, fd)

    def load_from_checkpoint(self) -> tuple[list[dict[str, Any | str]], int]:

        if self.dataset_name is None:
            self.dataset_name = str(int(time.time()))

        checkpoint_dir = os.path.join(os.path.dirname(__file__), "datasets")
        pattern = os.path.join(checkpoint_dir, f"{self.dataset_name}_batch_*.json")
        batch_files = sorted(
            glob.glob(pattern), key=lambda x: int(x.split("_")[-1].split(".")[0])
        )

        if not os.path.exists(os.path.join(checkpoint_dir, self.dataset_name)):
            return [], 0

        existing_outputs: list[dict[str, Any]] = []
        for batch_file in batch_files:
            with open(batch_file, "r") as fd:
                batch_data: list[dict[str, Any]] = json.load(fd)
                existing_outputs.extend(batch_data)

        # If there are batch files, set the next batch index to one higher than the last found batch index.
        if batch_files:
            last_batch_index = int(
                batch_files[-1].split("_")[-1].split(".")[0]
            )  # Assumes naming: datasetName_batch_N.json
            batch_index = last_batch_index + 1
        else:
            batch_index = 0

        return existing_outputs, batch_index

    def _call(
        self,
        inputs: Optional[dict[str, str] | None] = None,
    ) -> dict[str, List[dict[str, Any | str]]]:

        population = generate_population(few_shot_example_file=self.few_shot_file)

        batch_outputs: list[dict[str, Any]] = []
        outputs, batch_idx = self.load_from_checkpoint()

        while len(outputs) < self.k:
            few_shot = populate_few_shot(
                population=population, sample_size=self.sample_size
            )

            inputs = inputs or {}
            inputs["few_shot"] = few_shot
            generated_content: dict[str, str] = self.generator.invoke(input=inputs)

            for example in generated_content["generated"]:

                try:

                    normalizer_inputs: dict[str, Any] = {
                        self.normalizer.input_keys[0]: example
                    }
                    normalized_output = self.normalizer.invoke(normalizer_inputs)
                    output = {
                        "generated_input": example,
                        "normalized_output": normalized_output,
                    }
                    outputs.append(output)

                    if self.batch_save:
                        # Append the current output to the batch_outputs list
                        batch_outputs.append(output)
                        if len(batch_outputs) >= self.batch_size:
                            # Save the current batch to a file
                            self.save_batch(batch_outputs, batch_idx)
                            # Reset the batch_outputs list
                            batch_outputs = []
                            # Increment the batch index for the next batch
                            batch_idx += 1

                    if len(outputs) >= self.k:
                        break

                except Exception as e:
                    print(e)
                    continue

        # save any reminaing items to last batch
        if batch_outputs and self.batch_save:
            self.save_batch(batch_outputs, batch_idx)

        results: dict[str, dict[str, list[dict[str, Any | str]] | str]] = {
            "dataset": {
                "outputs": outputs,
                "generator_prompt": self.generator.template.template,
                "normalizer_prompt": self.normalizer.template.template,
            }
        }

        if self.manual_review:
            for pair in results["dataset"]["outputs"]:

                print(
                    f"{'Generated Input:', pair['generated_input']}\n-------\n{'Normalized Output:', pair['normalized_output']['normalized']}\n--------"
                )
                user_input = input("Accept? (y/N): ")
                if user_input.lower() == "y":
                    pair["manual_review"] = "accepted"
                else:
                    pair["manual_review"] = "rejected"

        if not self.batch_save:
            self.save_full_dataset(results)

        return results["dataset"]


def generate_dataset(
    generator_file: str,
    normalizer_file: str,
    example_file: str,
    k: int = 5,
    dataset_name: str | None = None,
    temperature: float = 0.8,
    cache: bool = False,
    manual_review: bool = False,
    model_name: str = "gpt-3.5-turbo",
    batch_save: bool = False,
    batch_size: int = 10,
):
    """Generate synthetic data and the normalized output

    Args:
        datatype (str): Name of the data to be generated, given by the template's name.
        k (int, optional): Number of samples to generate | Defaults to 10.
        dataset_name (str, optional): Name of generated data (user defined) | Defaults to None.
        temperature (float, optional): Parameter that affects the randomness of the output | Defaults to 0.8.
        cache (bool, optional): Determine wheher to cache calls to/from the API | Defaults to False.

    Returns:
        dict[str, dict[str, list[dict[str, Any | str]] | str]] : A dictionary of synthetically generated and normalized data
    """

    chain = DatasetPipeline.from_template(
        generator_template=generator_file,
        normalizer_template=normalizer_file,
        few_shot_file=example_file,
        k=k,
        temperature=temperature,
        cache=cache,
        model_name=model_name,
        batch_save=batch_save,
        batch_size=batch_size,
        dataset_name=dataset_name,
        manual_review=manual_review,
    )

    return chain.run()


if __name__ == "__main__":
    typer.run(generate_dataset)



import os
import logging 

logger = logging.getLogger("langchain")
logger.setLevel(logging.ERROR)

ROOT_DIR = os.path.dirname(os.path.dirname(__file__))
TEMPLATE_DIR = os.path.join(ROOT_DIR, "templates")
EXAMPLE_DIR = os.path.join(ROOT_DIR, "examples")



from transformers import TrainingArguments
import typer
from transformers import Trainer, AutoConfig
from fuxion.finetuning.dataset import tokenizer, NormalizationDataset
from transformers import DataCollatorForSeq2Seq
from transformers import T5ForConditionalGeneration
import mlflow

app = typer.Typer(pretty_exceptions_enable=False)


def test_generation(model, test_example):
    example = test_example["input_ids"].unsqueeze(0).to(model.device)
    output_tokens = model.generate(input_ids=example, max_length=256, temperature=0.0)
    output_text = tokenizer.batch_decode(output_tokens, skip_special_tokens=True)
    expected_output_text = tokenizer.decode(
        test_example["labels"], skip_special_tokens=True
    )
    print("Output", output_text)
    print("Expected Output", expected_output_text)


@app.command()
def main(
    experiment_name,
    data_file: str | None = None,
    dataset_size: int | None = None,
    dataloader_num_workers: int = 0,
    max_length: int = 512,
    batch_size: int = 2,
    dropout: float = 0.0,
    gradient_checkpointing: bool = False,
    pretrained_checkpoint: str | None = None,
    base_model: str = "google/byt5-small",
    num_train_epochs: float = 10.0,
    learning_rate: float = 3e-4,
    weight_decay: float = 0.01,
    warmup_ratio: float = 0.1,
    gradient_accumulation_steps: int = 1,
    # resume: bool=False,
    # max_steps: int = None,
):
    # TODO: start training from random initialization
    # TODO: incorporate other objectives
    model_cls = T5ForConditionalGeneration

    if pretrained_checkpoint is None:
        print("Training from pre-trained model")
        pretrained_checkpoint = base_model
        print(pretrained_checkpoint)

    config = AutoConfig.from_pretrained(pretrained_checkpoint)
    config.hidden_dropout_prob = dropout
    config.attention_probs_dropout_prob = dropout
    model = model_cls.from_pretrained(pretrained_checkpoint, config=config)

    args = TrainingArguments(
        output_dir=experiment_name,
        run_name=experiment_name,
        dataloader_num_workers=dataloader_num_workers,
        per_device_train_batch_size=batch_size,
        do_eval=False,
        evaluation_strategy="no",
        num_train_epochs=num_train_epochs,
        prediction_loss_only=False,
        gradient_accumulation_steps=gradient_accumulation_steps,
        ignore_data_skip=False,
        save_steps=1000,
        save_total_limit=2,
        save_strategy="steps",
        logging_steps=1,
        logging_first_step=True,
        learning_rate=learning_rate,
        warmup_ratio=warmup_ratio,
        weight_decay=weight_decay,
        # max_steps=max_steps,
        # report_to="wandb",
        gradient_checkpointing=gradient_checkpointing
        # fp16=True,
    )

    collator = DataCollatorForSeq2Seq(
        tokenizer=tokenizer,
        model=model,
        max_length=max_length,
        return_tensors="pt",
        padding="max_length",
    )

    train_dataset = NormalizationDataset(
        json_file=data_file, max_length=max_length, dataset_size=dataset_size
    )
    trainer_kwargs = dict(
        model=model,
        args=args,
        train_dataset=train_dataset,
        data_collator=collator,
    )
    trainer = Trainer(**trainer_kwargs)

    mlflow.autolog()
    trainer.train()
    test_generation(model, test_example=train_dataset[0])
    trainer.save_model()


if __name__ == "__main__":
    app()



from fuxion.normalizers import NormalizerChain

normalizer_chain = NormalizerChain.from_template(
    template_file="../templates/normalizer/address.template",
    temperature=0.0,
    cache=True,
    verbose=True,
    model_name="gpt-3.5-turbo",
)

normalizer = normalizer_chain.execute(
    example="John Doe street 1234, New York, NY 10001",
)
print(normalizer)


from typing import ClassVar
from langchain.chains.base import Chain
from typing import Any, Type
from langchain_community.cache import SQLiteCache
import langchain

langchain.llm_cache = SQLiteCache(database_path=".langchain.db")

class BaseChain(Chain):
    template_file: ClassVar[str]
    generator_template: ClassVar[str]
    normalizer_template: ClassVar[str]
    chain_type: ClassVar[str]
    registry: ClassVar[dict[Any, str]] = {}

    def __init_subclass__(cls, **kwargs: Any):
        super().__init_subclass__(**kwargs)
        cls.register(cls)

    @classmethod
    def register(cls, sub_cls: Any):
        if hasattr(sub_cls, "template_file"):
            cls.registry[(sub_cls.chain_type, sub_cls.template_file)] = sub_cls

    @classmethod
    def from_name(
        cls,
        template_file: str,
        class_suffix: str,
        base_cls: Type[Chain],
        *args,
        **kwargs
    ) -> Chain:

        template_name = template_file.split("/")[-1].split(".")[0]
        generated_type: type = type(
            template_name.capitalize() + class_suffix,
            (base_cls,),
            {"template_file": template_file},
        )
        return generated_type(*args, **kwargs)

    @classmethod
    def _from_name(
        cls,
        generator_template: str,
        normalizer_template: str,
        generator_chain: Chain,
        normalizer_chain: Chain,
        base_cls: Type[Chain],
        class_suffix: str,
        *args,
        **kwargs
    ) -> Chain:
        """ Used to generate dynamic classes for base class == DatasetPipeline

        Args:
            generator_template (str): _description_
            normalizer_template (str): _description_
            generator_chain (Chain): _description_
            normalizer_chain (Chain): _description_
            base_cls (Type[Chain]): _description_
            class_suffix (str): _description_

        Returns:
            Chain: _description_
        """
        template_name: str = generator_template.split("/")[-1].split(".")[0]

        if cls.chain_type != "DatasetPipeline":
            return
        else:
            generated_type: Type[Chain] = type(
                template_name.capitalize() + class_suffix,
                (base_cls,),
                {
                    "generator_template": generator_template,
                    "normalizer_template": normalizer_template,
                    "generator": generator_chain.from_name(
                        generator_template,
                        *args,
                        base_cls=generator_chain,
                        class_suffix="Generator",
                        **kwargs
                    ),
                    "normalizer": normalizer_chain.from_name(
                        normalizer_template,
                        *args,
                        base_cls=normalizer_chain,
                        class_suffix="Normalizer",
                        **kwargs
                    ),
                },
            )
            return generated_type(*args, **kwargs)



from transformers import TrainingArguments
import fire
import typer
from transformers import Trainer, AutoConfig
from fuxion.finetuning.dataset import tokenizer, NormalizationDataset
from transformers import DataCollatorForSeq2Seq
from transformers import T5ForConditionalGeneration
import wandb


def main(
    checkpoint: str,
    test_data_file: str,
    max_length: int = 512,
    batch_size: int = 4,
    dataset_size: int | None = None,
):
    model = T5ForConditionalGeneration.from_pretrained(checkpoint)

    collator = DataCollatorForSeq2Seq(
        tokenizer=tokenizer,
        model=model,
        max_length=max_length,
        return_tensors="pt",
        padding="max_length",
    )

    test_dataset = NormalizationDataset(
        json_file=test_data_file, max_length=max_length, dataset_size=dataset_size
    )

    results = {"correct": 0, "incorrect": 0}
    for batch_start in range(0, len(test_dataset), batch_size):
        # Making a batch to send to the model
        batch_end = min(batch_start + batch_size, len(test_dataset))
        examples = []
        for sample_idx in range(batch_start, batch_end):
            examples.append(test_dataset[sample_idx])

        # Using the collator to make sure all examples in our batch have the same shape
        # Padding and converting to torch tensors
        batch = collator(examples)

        # Running the model to produce a prediction for normalization
        output = model.generate(
            input_ids=batch["input_ids"], max_length=256, temperature=0.0
        )
        output_texts = tokenizer.batch_decode(output, skip_special_tokens=True)
        expected_texts = tokenizer.batch_decode(
            batch["labels"], skip_special_tokens=True
        )
        input_texts = tokenizer.batch_decode(
            batch["input_ids"], skip_special_tokens=True
        )
        for input_text, output_text, expected_text in zip(
            input_texts, output_texts, expected_texts
        ):
            if output_text.strip() == expected_text.strip():
                status = "correct"
            else:
                status = "incorrect"

            results[status] += 1
            print(
                f"Input: {input_text}\nOutput: {output_text}\nExpected: {expected_text}\nStatus: {status}\n-------"
            )
            acc = results["correct"] / (results["correct"] + results["incorrect"])
            print(
                f"\n\nCurrent Acc: {acc:0.2f} or {results['correct']}/{results['correct'] + results['incorrect']}\n\n"
            )


if __name__ == "__main__":
    typer.run(main)



import json
import ast
import re
from pprint import pprint
import typer
from pydantic import Field, PrivateAttr
from fuxion.base import BaseChain
from typing import Any, Optional
from langchain.chains import LLMChain
from langchain.prompts import PromptTemplate
from fuxion.models import get_model

optional_params = {"stop": ["]"]}
MODEL_CONFIG = {
    "gpt-4": optional_params,
    "gpt-3.5-turbo": optional_params,
}


class NormalizerChain(BaseChain):
    model_name: str = "gpt-3.5-turbo"
    template: PromptTemplate = PrivateAttr()
    chain = Field(LLMChain)
    chain_type = "normalizer"
    temperature: float = 0.0
    cache: bool = True
    verbose: bool = True
    datatype: Optional[str]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        with open(self.template_file) as f:
            data = f.read()
            pattern = r"\{\{(\w+)\}\}"
            match = re.search(pattern, data)
            self.datatype = match.group(1)

        self.template = PromptTemplate(
            input_variables=[self.datatype],
            template=open(self.template_file).read(),
            validate_template=True,
            template_format="jinja2",
        )
        if self.temperature > 2.0:
            raise ValueError(
                f"temperature:{self.temperature} is greater than the maximum of 2-'temperature'"
            )
        model_config = MODEL_CONFIG.get(self.model_name, {})
        if model_config:
            self.chain = LLMChain(
                prompt=self.template,
                llm=get_model(
                    model_name=self.model_name,
                    temperature=self.temperature,
                    cache=self.cache,
                    model_kwargs=model_config,
                ),
                verbose=self.verbose,
                )
        else:
            self.chain = LLMChain(
                prompt=self.template,
                llm=get_model(
                    model_name=self.model_name,
                    temperature=self.temperature,
                    cache=self.cache,
                ),
                verbose=self.verbose,
                )
    def execute(
        self,
        example: str,
    ):

        inputs = {self.datatype: example}
        result = self.chain.invoke(inputs)
        return result

    @classmethod
    def from_template(
        cls,
        *args,
        **kwargs,
    ):
        return super().from_name(
            *args,
            class_suffix="Normalizer",
            base_cls=NormalizerChain,
            chain_type="normalizer",
            **kwargs,
        )

    @property
    def input_keys(self) -> list[str]:
        return self.chain.input_keys

    @property
    def output_keys(self) -> list[str]:
        return ["normalized"]

    def _call(self, inputs: dict[str, str]) -> dict[str, list[Any]]:

        raw_output = self.chain.invoke(inputs)["text"]
        try:
            if self.model_name in MODEL_CONFIG:
                output = "[{" + self.chain.invoke(inputs)["text"] + "]"

                try:
                    return {"normalized": ast.literal_eval(output)}
                except json.JSONDecodeError:
                    print("Failed to normalize", output)
                    return {"normalized": []}
            else:
                cleaned_output = raw_output.lstrip('```python\n').rstrip('`')
                if not cleaned_output.strip().startswith('['):
                    cleaned_output = f"[{cleaned_output}]"

                normalized = json.loads(cleaned_output)
                if not isinstance(normalized, list):
                    normalized = [normalized]

                return {"normalized": normalized}
        except (json.JSONDecodeError, ValueError) as e:
            print("Failed to normalize", {e})
            print(f"Raw output: {raw_output}")
            return {"normalized": []}

def main(
    normalizer_template: str,
    example: str,
    temperature: float = 0.0,
    cache: bool = False,
    verbose: bool = True,
    model_name: str = "gpt-3.5-turbo",
):

    chain = NormalizerChain.from_template(
        normalizer_template,
        temperature=temperature,
        cache=cache,
        verbose=verbose,
        model_name=model_name,
    )
    if chain.datatype:
        result = chain.invoke({chain.datatype: example})
        pprint(result)


if __name__ == "__main__":
    typer.run(main)



import json
from torch.utils.data import Dataset
from transformers import AutoTokenizer
from typing import Any

tokenizer = AutoTokenizer.from_pretrained("google/byt5-small")
if tokenizer.pad_token is None:
    tokenizer.add_special_tokens({"pad_token": "[PAD]"})

def preprocess(example: dict[str, Any], max_length: int = 512) -> dict[str, Any]:

    encoded_inputs = tokenizer(
        example["generated_input"],
        padding="max_length",
        max_length=max_length,
        return_tensors="pt",
        truncation=True,
        return_overflowing_tokens=False,
    )

    encoded_outputs = tokenizer(
        json.dumps(example["normalized_output"]),
        padding="max_length",
        max_length=max_length,
        return_tensors="pt",
        truncation=True,
        return_overflowing_tokens=False,
    )
    result = {
        "input_ids": encoded_inputs.input_ids[0],
        "attention_mask": encoded_inputs.attention_mask[0],
        "labels": encoded_outputs.input_ids[0],
    }
    return result


class NormalizationDataset(Dataset):
    def __init__(
        self,
        json_file,
        max_length=2048,
        dataset_size=None,
    ):

        with open(json_file) as fd:
            self.dataset = json.load(fd)["dataset"]["outputs"]

        if dataset_size is not None:
            self.dataset = self.dataset[:dataset_size]
        self.max_length = max_length

    def __len__(self):
        return len(self.dataset)

    def __getitem__(self, index: int):
        example = self.dataset[index]
        processed_examples = preprocess(example, self.max_length)
        return processed_examples


if __name__ == "__main__":
    dataset = NormalizationDataset(
        json_file="/home/chief/fuxion/fuxion/datasets/qa1000.json",
    )
    print(dataset[15])



from fuxion.generators import GeneratorChain
from pprint import pprint

chain = GeneratorChain.from_template(
    template_file="examples/name_generator/generator.template",
    temperature=0.0,
    cache=True,
    verbose=True,
    model_name="gpt-3.5-turbo",
)


result = chain.execute(
    few_shot_example_file="examples/name_generator/few_shot.json", sample_size=3
)
pprint(result)



from typing import ClassVar
from pydantic import Field, PrivateAttr
from langchain.chains import LLMChain
from langchain.prompts import PromptTemplate
import typer
from pprint import pprint
from fuxion.base import BaseChain
from fuxion.few_shot import populate_few_shot, generate_population
from fuxion.settings import SEPARATOR
from fuxion.models import get_model

separator = SEPARATOR


class GeneratorChain(BaseChain):
    model_name: str = "gpt-3.5-turbo"
    template: PromptTemplate = PrivateAttr()
    chain = Field(LLMChain, required=False)
    chain_type: ClassVar[str] = "generator"
    temperature: float = 0.0
    cache: bool = False
    verbose: bool = True

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.template = PromptTemplate(
            input_variables=["few_shot"],
            template=open(self.template_file).read(),
            validate_template=False,
            template_format="jinja2",
        )
        if self.temperature > 2.0:
            raise ValueError(
                f"temperature:{self.temperature} is greater than the maximum of 2-'temperature'"
            )
        self.chain = LLMChain(
            prompt=self.template,
            llm=get_model(
                temperature=self.temperature,
                cache=self.cache,
                model_name=self.model_name,
            ),
            verbose=self.verbose,
        )

    def execute(
        self,
        few_shot_example_file: str,
        sample_size: int = 3,
    ):
        population = generate_population(few_shot_example_file=few_shot_example_file)
        few_shot = populate_few_shot(population=population, sample_size=sample_size)
        result = self.chain.invoke({"few_shot": few_shot})
        return result

    @classmethod
    def from_template(
        cls,
        *args,
        **kwargs,
    ):
        return super().from_name(
            *args,
            class_suffix="Generator",
            base_cls=GeneratorChain,
            chain_type="generator",
            **kwargs,
        )

    @property
    def input_keys(self) -> list[str]:
        return self.chain.input_keys

    @property
    def output_keys(self) -> list[str]:
        return ["generated"]

    def _call(self, inputs: dict[str, str]) -> dict[str, list[str]]:
        generated_items = self.chain.invoke(input=inputs)["text"].split(separator)
        filtered_items = [item for item in generated_items if item.strip()]
        return {"generated": filtered_items}


def main(
    generator_template: str,
    few_shot_example_file: str,
    sample_size: int = 3,
    temperature: float = 0.5,
    cache: bool = False,
    verbose: bool = True,
    model_name: str = "gpt-3.5-turbo",
):
    population = generate_population(few_shot_example_file=few_shot_example_file)
    few_shot = populate_few_shot(population=population, sample_size=sample_size)
    chain = GeneratorChain.from_template(
        generator_template,
        temperature=temperature,
        cache=cache,
        verbose=verbose,
        model_name=model_name,
    )
    pprint(chain.invoke(few_shot))


if __name__ == "__main__":
    typer.run(main)



from fuxion.pipelines import DatasetPipeline

irp_chain = DatasetPipeline.from_template(
    generator_template="irp/generator.template",
    normalizer_template="irp/normalizer.template",
    few_shot_file="irp/irp.json",
    dataset_name=f"irp_dataset_test_b",
    k=2,
    model_name="gpt-4o",
    cache=False,
    verbose=True,
    temperature=0.1,
)

result = irp_chain.execute()
print(result)



from langchain_openai import OpenAI, ChatOpenAI
from langchain_community.cache import SQLiteCache
from langchain.globals import set_llm_cache
import os

HUGGING_FACE_API_KEY = os.getenv("HUGGINGFACE_TOKEN")
OPENAI_API_KEY = os.getenv("OPENAI_API_KEY")
OPENAI_ORGANIZATION = os.getenv("OPENAI_ORGANIZATION")


MODEL_MAP = {
    "gpt-3.5-turbo": ChatOpenAI,
    "gpt-4": ChatOpenAI,
    "gpt-4o": ChatOpenAI,
    "gpt-4o-mini": ChatOpenAI,
}


def get_model(model_name: str, temperature: float, cache: bool, **kwargs) -> OpenAI:
    model_function = MODEL_MAP[model_name]

    if model_function is None:
        raise ValueError(f"Model {model_name} not found in model map, use one of {MODEL_MAP.keys()}")
    if cache:
        set_llm_cache(SQLiteCache(database_path=".langchain.db"))

    return model_function(
        model=model_name,
        temperature=temperature,
        cache=cache,
        api_key=OPENAI_API_KEY,
        organization=OPENAI_ORGANIZATION,
        **kwargs,
    )



from fuxion.pipelines import DatasetPipeline


pipeline_chain = DatasetPipeline.from_template(
    generator_template="examples/name_generator/generator.template",
    normalizer_template="examples/name_generator/normalizer.template",
    few_shot_file="examples/name_generator/few_shot.json",
    dataset_name="name_pipeline",
    k=20,
    model_name="gpt-3.5-turbo",
    cache=True,
    verbose=True,
    temperature=1.0,
    batch_save=True,
    batch_size = 3,
)

result = pipeline_chain.execute()
print(result)


import random
import os
import json
from typing import Union
from fuxion.settings import SEPARATOR

separator = SEPARATOR

def generate_population(few_shot_example_file: str) -> Union[list[str], list[dict[str, str]]]:
    population: Union[list[str], list[dict[str, str]]] = []

    if os.path.exists(few_shot_example_file):
        with open(few_shot_example_file, "r") as file:
            data = json.load(file)
        # Check if the JSON structure is a dictionary containing a list (e.g., under a key like "examples")
        if isinstance(data, dict) and "examples" in data:
            population = data["examples"]
        elif isinstance(data, list):
            population = data
    return population


def convert_to_string(example: Union[str, dict[str, str]]) -> str:
    if isinstance(example, str):
        return example
    elif isinstance(example, dict):
        new_examples = "\n".join(f"{key}: {value}" for key, value in example.items())
        return new_examples
    elif isinstance(example, list):
        # Process each dictionary in the list and join them with a separator
        dict_strings = ["\n".join(f"{key}: {value}" for key, value in d.items()) for d in example]
        return "\n---\n".join(dict_strings)


def populate_few_shot(
    population: Union[list[str], list[dict[str, str]]], sample_size: int
) -> str:
    sample = random.sample(population, k=sample_size)
    sample_text = separator.join([convert_to_string(example) for example in sample])
    few_shot = f"Reference Format:\n{sample_text}\n"
    return few_shot
